import pytest


@pytest.fixture(scope="module")
def service_params(service_params):
    service_params.update(backend_version="2")
    return service_params


def test_user_key(proxy, apicast_http_client):
    assert apicast_http_client.get("/get").status_code == 200



from tests.integration import asserts
from threescale_api.resources import Proxy, Service
from .asserts import assert_resource, assert_resource_params


def test_3scale_url_is_set(api, url, token):
    assert url is not None
    assert token is not None
    assert api.url is not None


def test_services_list(api):
    services = api.services.list()
    assert len(services) >= 1


def test_service_can_be_created(api, service_params, service):
    assert_resource(service)
    assert_resource_params(service, service_params)


def test_service_can_be_read(api, service_params, service):
    read = api.services.read(service.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, service_params)


def test_service_can_be_read_by_name(api, service_params, service):
    account_name = service['system_name']
    read = api.services[account_name]
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, service_params)


def test_service_can_be_updated(api, service):
    assert service['backend_version'] == '1'
    service['backend_version'] = '2'
    service.update()
    assert service['backend_version'] == '2'
    updated = service.read()
    assert updated['backend_version'] == '2'
    assert service['backend_version'] == '2'


def test_service_get_proxy(api, service: Service, proxy: Proxy, api_backend):
    assert proxy['api_backend'] == api_backend
    assert proxy['api_test_path'] == '/get'


def test_service_set_proxy(api, service: Service, proxy: Proxy, api_backend):
    updated = proxy.update(params=dict(api_test_path='/ip'))
    assert updated['api_backend'] == api_backend
    assert updated['api_test_path'] == '/ip'


def test_service_proxy_promote(service, proxy):
    res = proxy.promote()
    assert res is not None
    assert res['environment'] == 'production'
    assert res['content'] is not None


def test_service_proxy_deploy(service, proxy):
    # this will not propagate to proxy config but it allows deployment
    proxy.update(params=dict(support_email='test@example.com'))
    proxy.deploy()
    res = proxy.configs.list(env='staging')
    proxy_config = res.entity['proxy_configs'][-1]['proxy_config']
    assert proxy_config is not None
    assert proxy_config['environment'] == 'sandbox'
    assert proxy_config['content'] is not None
    assert proxy_config['version'] > 1


def test_service_list_configs(service, proxy):
    res = proxy.configs.list(env='staging')
    assert res
    item = res[0]
    assert item


def test_service_proxy_configs_version(service, proxy):
    config = service.proxy.list().configs.version(version=1)
    assert config
    assert config['environment'] == "sandbox"
    assert config['version'] == 1
    assert config['content']


def test_service_proxy_configs_latest(service, proxy):
    config = service.proxy.list().configs.latest()
    assert config
    assert config['environment'] == "sandbox"
    assert config['version']
    assert config['content']


def test_service_mapping_rules(service):
    map_rules = service.mapping_rules.list()
    assert len(map_rules) >= 1


def test_service_backend_usages_backend(backend_usage, backend):
    assert backend_usage.backend.entity_id == backend.entity_id



from tests.integration import asserts
from threescale_api.resources import Backends
from .asserts import assert_resource, assert_resource_params


def test_3scale_url_is_set(api, url, token):
    assert url is not None
    assert token is not None
    assert api.url is not None


def test_backends_list(api):
    backends = api.backends.list()
    assert len(backends) >= 1


def test_backend_can_be_created(api, backend_params, backend):
    assert_resource(backend)
    assert_resource_params(backend, backend_params)


def test_backend_can_be_read(api, backend_params, backend):
    read = api.backends.read(backend.entity_id)
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, backend_params)


def test_backend_can_be_read_by_name(api, backend_params, backend):
    backend_name = backend['system_name']
    read = api.backends[backend_name]
    asserts.assert_resource(read)
    asserts.assert_resource_params(read, backend_params)


def test_backend_can_be_updated(api, backend):
    assert backend['description'] == '111'
    backend['description'] = '222'
    backend.update()
    assert backend['description'] == '222'
    updated = backend.read()
    assert updated['description'] == '222'
    assert backend['description'] == '222'


def test_backend_metrics_list(backend, backend_metric):
    assert len(backend.metrics.list()) > 1


def test_backend_mapping_rules_list(backend, backend_mapping_rule):
    assert backend.mapping_rules.list()



import logging
import shlex
from typing import Union
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


def extract_response(response: requests.Response, entity: str = None,
                     collection: str = None) -> Union[dict, list]:
    """Extract the response from the response
    Args:
        response(requests.Response): Response
        entity(str): entity name to be extracted
        collection(str): collection name to be extracted
    Returns(Union[dict, list]): Extracted entity or list of entities
    """
    extracted: dict = response.json()
    if collection and collection in extracted:
        extracted = extracted.get(collection)
    if isinstance(extracted, list):
        return [value.get(entity) for value in extracted]
    if entity in extracted.keys():
        return extracted.get(entity)
    return extracted


class HttpClient:
    """3scale specific!!! HTTP Client

    This provides client to easily run api calls against provided service.
    Due to some delays in the infrastructure the client is configured to retry
    calls under certain conditions. To modify this behavior customized session
    has to be passed. session has to be fully configured in such case
    (e.g. including authentication"

    :param app: Application for which client should do the calls
    :param endpoint: either 'sandbox_endpoint' (staging) or 'endpoint' (production),
        defaults to sandbox_endpoint
    :param session: Used instead of default; it has to be fully configured
    :param verify: SSL verification
    """

    def __init__(self, app, endpoint: str = "sandbox_endpoint",
                 session: requests.Session = None, verify: bool = None):
        self._app = app
        self._endpoint = endpoint
        if session is None:
            session = requests.Session()
            self.retry_for_session(session)

            session.auth = app.authobj

        if verify is not None:
            session.verify = verify

        self._session = session

        logger.debug("[HTTP CLIENT] New instance: %s", self._base_url)

    @staticmethod
    def retry_for_session(session: requests.Session, total: int = 8):
        retry = Retry(
            total=total,
            backoff_factor=1,
            status_forcelist=(503, 404),
            raise_on_status=False,
            respect_retry_after_header=False
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)
        session.mount("http://", adapter)

    @property
    def _base_url(self) -> str:
        """Determine right url at runtime"""
        return self._app.service.proxy.fetch()[self._endpoint]

    def request(self, method, path,
                params=None, data=None, headers=None, cookies=None, files=None,
                auth=None, timeout=None, allow_redirects=True, proxies=None,
                hooks=None, stream=None, verify=None, cert=None, json=None) -> requests.Response:
        """mimics requests interface"""
        url = urljoin(self._base_url, path)

        req = requests.Request(
            method=method.upper(),
            url=url,
            headers=headers,
            files=files,
            data=data or {},
            json=json,
            params=params or {},
            auth=auth,
            cookies=cookies,
            hooks=hooks,
        )
        prep = self._session.prepare_request(req)

        logger.info("[CLIENT]: %s", request2curl(prep))

        send_kwargs = {
            "timeout": timeout,
            "allow_redirects": allow_redirects
        }

        proxies = proxies or {}

        send_kwargs.update(
            self._session.merge_environment_settings(prep.url, proxies, stream, verify, cert))

        response = self._session.send(prep, **send_kwargs)

        return response

    def get(self, *args, **kwargs) -> requests.Response:
        """mimics requests interface"""
        return self.request('GET', *args, **kwargs)

    def post(self, *args, **kwargs) -> requests.Response:
        """mimics requests interface"""
        return self.request('POST', *args, **kwargs)

    def patch(self, *args, **kwargs) -> requests.Response:
        """mimics requests interface"""
        return self.request('PATCH', *args, **kwargs)

    def put(self, *args, **kwargs) -> requests.Response:
        """mimics requests interface"""
        return self.request('PUT', *args, **kwargs)

    def delete(self, *args, **kwargs) -> requests.Response:
        """mimics requests interface"""
        return self.request('DELETE', *args, **kwargs)


def request2curl(request: requests.PreparedRequest) -> str:
    """Create curl command corresponding to given request"""

    cmd = ["curl", "-X %s" % shlex.quote(request.method)]
    if request.headers:
        cmd.extend([
            "-H %s" % shlex.quote(f"{key}: {value}")
            for key, value in request.headers.items()])
    if request.body:
        body = request.body
        if isinstance(body, bytes):
            body = body.decode("utf-8")
        cmd.append("-d %s" % shlex.quote(body))
    cmd.append(shlex.quote(request.url))

    return " ".join(cmd)
